# -*- coding: utf-8 -*-

from bokeh import palettes


def colors(num_colors):
    if num_colors < 3:
        return iter(palettes.Set1[3][:num_colors])

    else:
        try:
            return iter(palettes.Set1[num_colors])

        except KeyError:
            raise Exception(
                "Too many functions to benchmark, we only have colors to "
                "support {}".format(max(palettes.Set1.keys()))
            )



# -*- coding: utf-8 -*-

"""Main module."""

import collections
import time
import uuid

from bokeh import io as bi
from bokeh import models as bm
import bokeh.palettes
from bokeh import plotting as bp
from IPython import display as ipdisplay
import numpy as np
import pandas as pd

from perfume import analyze
from perfume import colors


class Timer(object):

    def __enter__(self):
        self._begin = time.perf_counter()
        return self

    def __exit__(self, *exc_info):
        self._end = time.perf_counter()

    @property
    def begin(self):
        return self._begin

    @property
    def end(self):
        return self._end

    def elapsed_seconds(self):
        return self.end - self.begin

    @classmethod
    def time(cls, fn, *args, **kwargs):
        with cls() as timer:
            fn(*args, **kwargs)
        return timer.elapsed_seconds()


class Display(object):

    def __init__(self, names, initial_size, width=900, height=480):
        # Call this once to raise an error early if necessary:
        self._colors = colors.colors(len(names))

        self._start = time.perf_counter()
        self._initial_size = initial_size
        self._sources = collections.OrderedDict(
            [
                (
                    name,
                    {
                        "hist": bm.ColumnDataSource(
                            data={"top": [], "left": [], "right": []}
                        ),
                        "pdf": bm.ColumnDataSource(data={"x": [], "y": []}),
                        "stddev": bm.ColumnDataSource(
                            data={"base": [], "lower": [], "upper": []}
                        ),
                        "median": bm.ColumnDataSource(data={"x": [], "y": []}),
                    },
                )
                for name in names
            ]
        )
        self._width = width
        self._height = height
        self._plot = None
        self._elapsed_rendering_seconds = 0.0
        self._describe_widget = ipdisplay.HTML("")
        self._display_id = str(uuid.uuid1())

    def elapsed_rendering_ratio(self):
        elapsed = time.perf_counter() - self._start
        return self._elapsed_rendering_seconds / elapsed

    def initialize_plot(self, title):
        with Timer() as timer:
            plot = bp.figure(
                title=title, plot_width=self._width, plot_height=self._height
            )
            plot.xaxis.axis_label = "millis"
            plot.yaxis.visible = False
            _colors = iter(self._colors)
            for name, sources in self._sources.items():
                color = next(_colors)
                plot.quad(
                    top="top",
                    bottom=0,
                    left="left",
                    right="right",
                    source=sources["hist"],
                    alpha=0.3,
                    fill_color=color,
                    line_color=color,
                )
                plot.line(
                    "x",
                    "y",
                    source=sources["pdf"],
                    legend=name,
                    alpha=0.5,
                    line_color=color,
                    line_width=4,
                )
                stddev = bm.Whisker(
                    base="base",
                    lower="lower",
                    upper="upper",
                    source=sources["stddev"],
                    dimension="width",
                    line_alpha=0.7,
                    line_color=color,
                    line_width=2,
                )
                for head in (stddev.lower_head, stddev.upper_head):
                    head.line_color = color
                    head.line_width = 2
                    head.line_alpha = 0.7
                plot.add_layout(stddev)
                median = bm.Whisker(
                    base="y",
                    lower="x",
                    upper="x",
                    source=sources["median"],
                    dimension="width",
                    line_alpha=0.7,
                    line_color=color,
                    line_width=2,
                )
                for head in (median.lower_head, median.upper_head):
                    head.line_color = color
                    head.line_width = 2
                    head.line_alpha = 0.7
                plot.add_layout(median)

        self._elapsed_rendering_seconds -= timer.elapsed_seconds()
        return plot

    @staticmethod
    def _ks_style(s):
        if np.isnan(s):
            return "visibility: hidden"

        else:
            thresholds = [1.22, 1.36, 1.48, 1.63, 1.73, 1.95]
            cs = list(reversed(bokeh.palettes.RdYlGn[len(thresholds) + 1]))
            color = cs[np.searchsorted(thresholds, s)]
            return "background-color: {}".format(color)

    def update(self, samples):
        # If this is a module-level import, readthedocs fails because
        # this triggers an import of _tkinter, which isn't built in to
        # the python that they use.
        import seaborn as sns

        with Timer() as timer:
            timings = analyze.timings(samples)
            bucketed_timings = analyze.bucket_resample_timings(samples)
            for name, sources in self._sources.items():
                array = timings[name].values
                hist, edges = np.histogram(array, density=True, bins="auto")
                x, y = sns.distributions._statsmodels_univariate_kde(
                    array,
                    "gau",
                    "scott",
                    200,
                    3,
                    (-np.inf, np.inf),
                    cumulative=False,
                )
                whisker_height = np.max(y) / 2
                lower, median, upper = np.percentile(array, [25., 50., 75.])

                sources["hist"].data = {
                    "top": hist, "left": edges[:-1], "right": edges[1:]
                }
                sources["pdf"].data = {"x": x, "y": y}
                sources["stddev"].data = {
                    "base": [whisker_height],
                    "lower": [lower],
                    "upper": [upper],
                }
                sources["median"].data = {"x": [median], "y": [whisker_height]}

            describe_html = (
                timings.describe().style.set_precision(3).set_caption(
                    "Descriptive Timing Statistics"
                ).render()
            )
            if len(self._sources) > 1:
                ks_frame = analyze.ks_test(timings)
                ks_bk_frame = analyze.ks_test(bucketed_timings)
                ks_html = (
                    ks_frame.style.applymap(self._ks_style).set_precision(
                        3
                    ).set_caption(
                        "K-S test"
                    ).render()
                )
                ks_bk_html = (
                    ks_bk_frame.style.applymap(self._ks_style).set_precision(
                        2
                    ).set_caption(
                        "Bucketed K-S test"
                    ).render()
                )
                html = describe_html + ks_html + ks_bk_html
                self._describe_widget.data = html.replace(
                    "table", 'table style="display:inline"'
                )
            else:
                self._describe_widget.data = describe_html

            total_bench_time = timings[self._initial_size:].sum().sum() / 1000.
            elapsed = time.perf_counter() - self._start
            num_samples = len(timings.index)
            title = (
                "{} samples, {:.2f} sec elapsed, {:.2f} samples/sec, "
                "{:.2f}% efficiency"
            ).format(
                num_samples,
                elapsed,
                (num_samples - self._initial_size) / elapsed,
                100. * total_bench_time / elapsed
            )

            if self._plot is None:
                self._plot = self.initialize_plot(title)
                bi.show(self._plot, notebook_handle=True)
                ipdisplay.display(
                    self._describe_widget, display_id=self._display_id
                )
            else:
                self._plot.title.text = title
                bi.push_notebook()
                ipdisplay.update_display(
                    self._describe_widget, display_id=self._display_id
                )
        self._elapsed_rendering_seconds += timer.elapsed_seconds()


def _flatten(l):
    return [n for sublist in l for n in sublist]


def bench(*fns, samples=None, efficiency=.9):
    """Benchmarks functions, displaying results in a Jupyter notebook.

    Runs ``fns`` repeatedly, collecting timing information, until
    :exc:`KeyboardInterrupt` is raised, at which point benchmarking
    stops and the results so far are returned.

    Parameters
    ----------
    fns : list of callable
        A list of functions to benchmark and compare
    samples : pandas.DataFrame
        Optionally, pass the results of a previous call to
        :func:`.bench` to continue from its already collected data.
    efficiency : float
        Number between 0 and 1.  Represents the target portion of time
        we aim to spend running the functions under test (so, we spend
        up to :math:`1 - efficiency` time analyzing and rendering
        plots).

    Returns
    -------
    pandas.DataFrame
        A dataframe containing the results so far.  The row index is
        just an autoincrement integer, and the column index is a
        :class:`~pandas.MultiIndex` where the first level is function
        name and the second level is ``begin`` or ``end``.
    """
    if samples is None:
        sample_records = []
    else:
        sample_records = [tuple(r) for r in samples.to_records(index=False)]
    names = [fn.__name__ for fn in fns]
    disp = Display(names, len(sample_records))
    index = pd.MultiIndex(
        levels=[names, ("begin", "end")],
        labels=[
            _flatten([(i, i) for i in range(len(names))]), [0, 1] * len(names)
        ],
        names=("function", "timing"),
    )
    try:
        while True:
            sample = []
            for fn in fns:
                with Timer() as timer:
                    fn()
                sample.extend((timer.begin, timer.end))
            sample_records.append(tuple(t * 1000 for t in sample))

            if (
                len(sample_records) > 10
                and disp.elapsed_rendering_ratio() < (1. - efficiency)
            ):
                samples = pd.DataFrame.from_records(
                    iter(sample_records), columns=index
                )
                disp.update(samples)
    except KeyboardInterrupt:
        return pd.DataFrame.from_records(iter(sample_records), columns=index)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `perfume` package.

perfume is fairly visualization-heavy and deals with stochastic
events, so end-to-end testing isn't really aimed for here.  But we can
test the transformations in analyze somewhat.
"""


import unittest

import numpy as np
import numpy.testing as npt
import pandas as pd
import pandas.util.testing as pdt

from perfume import analyze


class TestAnalyze(unittest.TestCase):
    """Tests for `perfume.analyze` module."""

    def setUp(self):
        samples = []
        t = 1.0
        for i in range(20):
            sample = []
            sample.append(t)
            t += 1.1
            sample.append(t)
            t += 0.2
            sample.append(t)
            t += 1.5
            sample.append(t)
            t += 0.1
            samples.append(sample)
        self.samples = pd.DataFrame(
            data=samples,
            columns=pd.MultiIndex(
                levels=[["fn1", "fn2"], ["begin", "end"]],
                labels=[[0, 0, 1, 1], [0, 1, 0, 1]],
            ),
        )

    def tearDown(self):
        del self.samples

    def test_timings(self):
        """Test that timings gives us the right results."""
        timings = analyze.timings(self.samples)
        pdt.assert_index_equal(timings.index, self.samples.index)
        self.assertSetEqual(
            set(timings.columns), set(self.samples.columns.get_level_values(0))
        )
        self.assertEqual(len(timings.columns), len(self.samples.columns) / 2)
        npt.assert_array_almost_equal(timings["fn1"], 1.1)
        npt.assert_array_almost_equal(timings["fn2"], 1.5)

    def test_isolate(self):
        """Test that isolate gives us the right results."""
        isolated = analyze.isolate(self.samples)
        pdt.assert_index_equal(isolated.index, self.samples.index)
        pdt.assert_index_equal(isolated.columns, self.samples.columns)
        pdt.assert_frame_equal(
            analyze.timings(isolated), analyze.timings(self.samples)
        )
        npt.assert_array_almost_equal(
            isolated["fn1"]["begin"], np.arange(20) * 1.1
        )
        npt.assert_array_almost_equal(
            isolated["fn1"]["end"], 1.1 + (np.arange(20) * 1.1)
        )
        npt.assert_array_almost_equal(
            isolated["fn2"]["begin"], np.arange(20) * 1.5
        )
        npt.assert_array_almost_equal(
            isolated["fn2"]["end"], 1.5 + (np.arange(20) * 1.5)
        )

    def test_timings_in_context(self):
        """Test that timings_in_context gives us the right results."""
        in_context = analyze.timings_in_context(self.samples)
        # Since each "function" has a fixed frequency, we can create
        # two series with TimedeltaIndexes and align them into the
        # same DataFrame, which should be what timings_in_context
        # gives us.
        fn1_expected = pd.Series(
            1.1,
            index=pd.TimedeltaIndex(
                freq=pd.Timedelta("1.1s"),
                start="1.1s",
                periods=20,
                name="time",
            ),
        )
        fn2_expected = pd.Series(
            1.5,
            index=pd.TimedeltaIndex(
                freq=pd.Timedelta("1.5s"),
                start="1.5s",
                periods=20,
                name="time",
            ),
        )
        expected = pd.DataFrame({"fn1": fn1_expected, "fn2": fn2_expected})
        pdt.assert_frame_equal(in_context, expected)



# -*- coding: utf-8 -*-

""":mod:`perfume.analyze` contains transformation and analysis tools.

These functions mostly take as input the samples collected by
:func:`perfume.bench`.
"""

import bokeh.io as bi
import bokeh.models as bm
import bokeh.plotting as bp
import numpy as np
import pandas as pd
from scipy import stats

from perfume import colors


def timings(samples):
    """Converts samples to sample times per observation."""
    groups = samples.groupby(axis=1, level=0)
    return groups.apply(lambda group: group.iloc[:, 1] - group.iloc[:, 0])


def _remove_other_timings(group):
    other_timings = (
        (group.iloc[:, 0] - group.iloc[:, 1].shift(1)).fillna(0).cumsum()
    )
    ret = group.groupby(axis=1, level=1).apply(
        lambda x: x.iloc[:, 0] - other_timings
    )
    ret.columns = group.columns
    return ret


def isolate(samples):
    """For each function, isolates its begin and end times.

    Within each function's begins and ends, each begin will be equal
    to the previous end.  This gives a sequence of begins and ends as
    if each function were run in isolation with no benchmarking
    overhead.
    """
    zeroed = samples.groupby(axis=1, level=0).apply(
        lambda group: group - group.iloc[0, 0]
    )
    return zeroed.groupby(axis=1, level=0).apply(_remove_other_timings)


def _timing_in_context(group):
    time_col = pd.DataFrame(data=group.iloc[:, 1].values, columns=["time"])
    values = pd.DataFrame(data=timings(group), columns=[group.columns[0][0]])
    time_col = pd.to_timedelta(time_col["time"], unit="s")
    values = values.assign(time=time_col)
    return values.set_index("time")


def timings_in_context(samples):
    """Returns a sparse dataframe with a time index, with timings.

    Each cell contains the timing observed, at the time when it was
    observed.  Therefore, each row will have NaNs except for the
    function whose sample completed at that time.
    """
    iso = isolate(samples)
    t = iso.groupby(axis=1, level=0).apply(_timing_in_context)
    t.columns = t.columns.droplevel(0)
    return t


def bucket_resample_timings(
    samples, sample_size=10, agg=np.mean, sample_count=1000
):

    def _meat_axe(s):
        return pd.Series(
            [
                agg(np.random.choice(s.values, size=sample_size, replace=True))
                for _ in range(sample_count)
            ]
        )

    return timings(samples).apply(_meat_axe)


def _ks_Z(a, b):
    result = stats.ks_2samp(a, b)
    n = len(a)
    m = len(b)
    return result.statistic / np.sqrt((n + m) / (n * m))


def ks_test(t):
    """Runs the Kolmogorov-Smirnov test across functions.

    Returns a DataFrame containing all pairwise K-S test results.

    The standard K-S test computes :math:`D`, which is the maximum
    difference between the empirical CDFs.

    The result value we return is the :math:`Z` value, defined as

    .. math::

       Z = D / \\sqrt{(n + m) / nm}

    where :math:`n` and :math:`m` are the respective sample sizes.

    :math:`Z` is typically interpreted using a lookup table, i.e. for
    confidence level :math:`\\alpha`, we want to see a :math:`Z`
    greater than :math:`c(\\alpha)`:

    +--------------------+------+------+-------+------+-------+-------+
    | :math:`\\alpha`     | 0.10 | 0.05 | 0.025 | 0.01 | 0.005 | 0.001 |
    +--------------------+------+------+-------+------+-------+-------+
    | :math:`c(\\alpha)`  | 1.22 | 1.36 | 1.48  | 1.63 | 1.73  | 1.95  |
    +--------------------+------+------+-------+------+-------+-------+
    """
    data = {
        name: (
            [
                _ks_Z(t[name].values, t[t.columns[j]].values)
                for j in range(i + 1)
            ]
            + ([np.nan] * (len(t.columns) - 2 - i))
        )
        for i, name in enumerate(t.columns[1:])
    }
    idx = pd.Index(t.columns[:-1], name="K-S test Z")
    return pd.DataFrame(data, index=idx)


def _cumulative_quantiles(group, rng):
    group = isolate(group)
    t = timings(group)
    ret = pd.concat(
        [
            t.iloc[:(i + 1), :].describe().T[
                ["min", "25%", "50%", "75%", "max"]
            ]
            for i in rng
        ]
    )
    ret = ret.set_index(group.iloc[slice(rng.start, rng.stop), 1])
    ret.index.name = "time"
    return ret


def cumulative_quantiles(samples, rng=None):
    """Computes "cumulative quantiles" for each function.

    That is, for each time, what are the extremes, median, and
    25th/75th percentiles for all observations up until that point.
    """
    if rng is None:
        rng = range(len(samples.index))
    groups = samples.groupby(axis=1, level=0)
    df = groups.apply(lambda group: _cumulative_quantiles(group, rng))
    return df


def cumulative_quantiles_plot(
    samples, plot_width=960, plot_height=480, show_samples=True
):
    """Plots the cumulative quantiles along with a scatter plot of
    observations."""
    plot = bp.figure(plot_width=960, plot_height=480)

    names = samples.columns.levels[0]
    _colors = {
        name: color for name, color in zip(names, colors.colors(len(names)))
    }

    def draw(group):
        name = group.columns[0][0]
        color = _colors[name]
        group.columns = group.columns.droplevel(0)
        group = group.dropna()
        quantile_source = bm.ColumnDataSource(
            pd.DataFrame(
                data={"lower": group["25%"], "upper": group["75%"]},
                index=group.index,
            ).dropna().reset_index()
        )
        extreme_source = bm.ColumnDataSource(
            pd.DataFrame(
                data={"lower": group["min"], "upper": group["max"]},
                index=group.index,
            ).dropna().reset_index()
        )
        plot.line(group.index, group["50%"], line_color=color, legend=name)
        plot.add_layout(
            bm.Band(
                base="time",
                lower="lower",
                upper="upper",
                source=quantile_source,
                fill_color=color,
                fill_alpha=0.2,
            )
        )
        plot.add_layout(
            bm.Band(
                base="time",
                lower="lower",
                upper="upper",
                source=extreme_source,
                fill_color=color,
                fill_alpha=0.025,
            )
        )
        plot.line(
            "time", "lower", line_color=color, alpha=0.5, source=extreme_source
        )
        plot.line(
            "time", "upper", line_color=color, alpha=0.5, source=extreme_source
        )

    cumulative_quantiles(samples).groupby(axis=1, level=0).apply(draw)

    if show_samples:

        def scatter(group):
            name = group.columns[0][0]
            color = _colors[name]
            group = isolate(group)
            t = timings(group).set_index(group.iloc[:, 1])
            t.index.name = "time"
            t.columns = ["value"]
            source = bm.ColumnDataSource(t.reset_index())
            plot.circle(
                x="time",
                y="value",
                source=source,
                color=color,
                size=1,
                alpha=0.5,
            )

        samples.groupby(axis=1, level=0).apply(scatter)

    bi.show(plot)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages
import versioneer

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    'bokeh>=0.12',
    'ipython>=5.0',
    'ipywidgets>=5.0',
    'matplotlib>=2.0',
    'notebook>=5.0',
    'numpy>=1.11',
    'pandas>=0.19',
    'seaborn>=0.7',
    'statsmodels>=0.8',
]

setup_requirements = [
    # TODO(leifwalsh): put setup requirements (distutils extensions, etc.) here
]

test_requirements = [
    # TODO: put package test requirements here
]

setup(
    name='perfume-bench',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="Interactive performance benchmarking in Jupyter",
    long_description=readme + '\n\n' + history,
    author="Leif Walsh",
    author_email='leif.walsh@gmail.com',
    url='https://github.com/leifwalsh/perfume',
    packages=find_packages(include=['perfume']),
    entry_points={
        'console_scripts': [
            'perfume=perfume.cli:main'
        ]
    },
    include_package_data=True,
    install_requires=requirements,
    python_requires='>=3.4, <4',
    license="BSD license",
    zip_safe=False,
    keywords='perfume python performance benchmarking jupyter interactive',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    test_suite='tests',
    tests_require=test_requirements,
    setup_requires=setup_requirements,
)



# -*- coding: utf-8 -*-

"""Top-level package for perfume."""

from .perfume import bench  # noqa: F401
from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions

__author__ = """Leif Walsh"""
__email__ = "leif.walsh@gmail.com"
__all__ = ["bench"]
